
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include "window.h"
#include "placement.h"

#define TRUE  1
#define FALSE 0

void exit_program(Fl_Button *, void *) 
{
  exit ( 0 ) ;
}


void clrCell ( int x, int y )
{
  assert ( x >= 0 && x <= 15 ) ;
  assert ( y >= 0 && y <= 15 ) ;

  letter_button [ x + y * 16 ] -> clear () ;
}


char setCell ( int x, int y )
{
  assert ( x >= 0 && x <= 15 ) ;
  assert ( y >= 0 && y <= 15 ) ;

  letter_button [ x + y * 16 ] -> set () ;

  return letter_button [ x + y * 16 ] -> label()[0] ;
}


void highlight ( int x, int y )
{
  assert ( x >= 0 && x <= 15 ) ;
  assert ( y >= 0 && y <= 15 ) ;

  letter_button [ x + y * 16 ] -> labelfont ( 1 ) ;
}


void clearAll ()
{
  for ( int i = 0 ; i < 256 ; i++ )
    letter_button [ i ] -> clear () ;
}


void letter (Fl_Button *, void *wh ) 
{
  static int first_letter = TRUE ;
  static int fx  = 0 ;
  static int fy  = 0 ;
  int which = *(int*)wh ;

  if ( game_over )
  {
    clearAll () ;
    return ;
  }

  if ( first_letter )
  {
    clearAll () ;  /* Just to be sure */
    fx =  which     & 0x0F ;
    fy = (which>>4) & 0x0F ;
    first_letter = FALSE ;
    setCell ( fx, fy ) ;
    setMessage ( "Select end of word..." ) ;
    return ;
  }

  /*
    Is this a legal screen coordinate?
  */

  int sx =  which     & 0x0F ;
  int sy = (which>>4) & 0x0F ;

  if ( fx == sx && fy == sy )  /* Same location */
  {
    setMessage ( "Oops!" ) ;
    clearAll () ;  /* Just to be sure */
    first_letter = TRUE ;
    return ;
  }

  /* Illegal move?   Just ignore it. */

  if ( ! ( ( fx == sx && fy != sy ) ||  /* Vertical   */
           ( fx != sx && fy == sy ) ||  /* Horizontal */
           ( fx-sx == fy-sy )       ||  /* Diagonal   */
           ( fx-sx == sy-fy ) ) )       /* Diagonal   */
  {
    setMessage ( "Illegal move!" ) ;
    clrCell ( sx, sy ) ;
    return ;
  }

  char word [ 17 ] ;
  int  n = 0 ;
  int  i ;

  word [ 0 ] = '\0' ;

  if ( fx == sx && fy != sy ) /* Vertical   */
  {
    for ( i = fy ; i != sy ; i += (fy<sy)?1:-1 )
      word[n++] = setCell ( fx, i ) ; 
    word[n++] = setCell ( fx, i ) ; 
  }
  else
  if ( fx != sx && fy == sy )  /* Horizontal */
  {
    for ( i = fx ; i != sx ; i += (fx<sx)?1:-1 )
      word[n++] = setCell ( i, fy ) ; 
    word[n++] = setCell ( i, fy ) ; 
  }
  else
  if ( fx-sx == fy-sy )
  {
    for ( i = 0 ; i != sx-fx ; i += (fx<sx)?1:-1 )
      word[n++] = setCell ( i+fx, i+fy ) ; 
    word[n++] = setCell ( i+fx, i+fy ) ; 
  }
  else
  if ( fx-sx == sy-fy )
  {
    for ( i = 0 ; i != sx-fx ; i += (sx>fx)?1:-1 )
      word[n++] = setCell ( i+fx, fy-i ) ; 
    word[n++] = setCell ( i+fx, fy-i ) ; 
  }
  else
  {
    printf ( "Eh?  fxy=(%d,%d),sxy=(%d,%d)\n", fx,fy,sx,sy ) ;
    assert ( FALSE ) ;
  }

  word [ n ] = '\0' ;

  int found = FALSE ;

  for ( int i = 0 ; i < next_word ; i++ )
  {
    if ( strcmp ( word, words[i] ) == 0 )
    {
      words[i][0] = '\0' ;
      updateWordlist () ;
      found = TRUE ;
      break ;
    }
  }

  first_letter = TRUE ;

  if ( ! found )
  {
    setMessage ( "That's not a word!" ) ;
    return ;
  }

  setMessage ( word ) ;

  if ( fx == sx && fy != sy ) /* Vertical   */
  {
    for ( i = fy ; i != sy ; i += (fy<sy)?1:-1 )
      highlight ( fx, i ) ; 
    highlight ( fx, i ) ; 
  }
  else
  if ( fx != sx && fy == sy )  /* Horizontal */
  {
    for ( i = fx ; i != sx ; i += (fx<sx)?1:-1 )
      highlight ( i, fy ) ; 
    highlight ( i, fy ) ; 
  }
  else
  if ( fx-sx == fy-sy )
  {
    for ( i = 0 ; i != sx-fx ; i += (fx<sx)?1:-1 )
      highlight ( i+fx, i+fy ) ; 
    highlight ( i+fx, i+fy ) ; 
  }
  else
  if ( fx-sx == sy-fy )
  {
    for ( i = 0 ; i != sx-fx ; i += (sx>fx)?1:-1 )
      highlight ( i+fx, fy-i ) ; 
    highlight ( i+fx, fy-i ) ; 
  }
}



void fetchWords ( char *wordlist, char *wordnext, char *words[] )
{
  int next = 0 ;

  /*
    'wordnext' is the file that tells us which puzzle to generate
    next.

    Open it, read the number - write it out with the next number up.
  */

  FILE *fd = fopen ( wordnext, "ra" ) ;
  if ( fd != NULL )
  {
    fscanf ( fd, "%d", &next ) ;
    fclose ( fd ) ;
  }

  fd = fopen ( wordnext, "wa" ) ;

  if ( fd != NULL )
  {
    fprintf ( fd, "%d <----This is the next puzzle you'll get.\n", next+1 ) ;
    fclose ( fd ) ;
  }
  else
  {
    fprintf ( stderr, "Failed to write to '%s'\n", wordnext ) ;
    next = rand () % 20 ;
  }

  fprintf ( stderr, "====================\n" ) ;
  fprintf ( stderr, "  Puzzle Number %d\n", next ) ;
  fprintf ( stderr, "====================\n" ) ;

  /****************************************************************/

  /*
    'wordlist' contains the list of words - one per line,
    separated by rows of '----------'.

    Parse the word file.
  */

  fd = fopen ( wordlist, "ra" ) ;

  if ( fd == NULL )
  {
    fprintf ( stderr, "File '%s' is missing.\n", wordlist ) ;
    exit ( 1 ) ;
  }

  char s [ 100 ] ;
  int i ;

  /* Skip the first 'next' puzzles */

  for ( i = 0 ; i < next ; i++ )
  {

    do
    {
      /*
        If we run off the end of the file then
        we need to reset the 'next puzzle' number to
        zero and try again.
      */

      if ( fgets ( s, 100, fd ) == NULL )
      {
        fclose ( fd ) ;

        next = 0 ;

	fd = fopen ( wordnext, "wa" ) ;

	if ( fd != NULL )
	{
	  fprintf ( fd, "%d <----This is the next puzzle you'll get.\n", 0 ) ;
	  fclose ( fd ) ;
	}
	else
	{
	  fprintf ( stderr, "Failed to write to '%s'\n", wordnext ) ;
	  exit ( 1 ) ;
	}

        /* Try again! */

        fetchWords ( wordlist, wordnext, words ) ;
        return ;
      }
    } while ( *s != '-' ) ;
  }

  for ( i = 0 ; ; i++ )
  {
    /*
      Read the words.
    */

    if ( fgets ( s, 100, fd ) == NULL || *s == '-' )
      break ;


    strcpy ( &(words [ i ][ 0 ]), s ) ;
    words [ i ][ strlen(words[i]) - 1 ] = '\0' ;
  }
  
  next_word = i ;
  words [ i ][ 0 ] = '\0' ;

  fclose ( fd ) ;
}



int main(int argc, char **argv)
{
  srand ( time ( NULL ) ) ;
  Fl_Window *w = make_window () ;

  next_word = 0 ;

  words = new char * [ 100 ] ;

  for ( int i = 0 ; i < 100 ; i++ )
    words[i] = new char [ 17 ] ;

  fetchWords ( ".wordsearch_puzzles", ".wordsearch_next", words ) ;

  updateWordlist () ;
  placement () ;

  w -> show ( argc, argv ) ;

  setMessage ( "by Steve Baker." ) ;
  return Fl::run () ;
}


